# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).

from airtk.defense.feature_scatter.fs_train import (
    FeatureScatterTraining,
)

__all__ = ["FeatureScatterTraining"]
