# python adt_training.py --epochs 1
# python clean_training.py --epochs 1
# python custat_training.py --epochs 1
# python standardat_training.py --epochs 1
# python3 testing_framework.py --epochs 1
python train_va.py --epochs 1
# python train.py --epochs 1
# python3 warmup_round.py --epochs 1